@@ -1,8 +1,8 @@
 """
 Translation support for admin forms.
 """
+from __future__ import unicode_literals
 import django
-from django.conf import settings
 from django.conf.urls import patterns, url
 from django.contrib import admin
 from django.contrib.admin.options import csrf_protect_m, BaseModelAdmin, InlineModelAdmin
@@ -11,18 +11,20 @@ from django.core.exceptions import PermissionDenied, ImproperlyConfigured
 from django.core.urlresolvers import reverse
 from django.db import router
 from django.forms import Media
-from django.forms.models import BaseInlineFormSet
 from django.http import HttpResponseRedirect, Http404, HttpRequest
 from django.shortcuts import render
-from django.utils.encoding import iri_to_uri, force_unicode
+from django.utils.encoding import iri_to_uri, force_text
 from django.utils.functional import lazy
-from django.utils.translation import ugettext_lazy as _
+from django.utils.http import urlencode
+from django.utils.translation import ugettext_lazy as _, get_language
+from django.utils import six
 from parler import appsettings
-from parler.forms import TranslatableModelForm
+from parler.forms import TranslatableModelForm, TranslatableBaseInlineFormSet
 from parler.managers import TranslatableQuerySet
 from parler.models import TranslatableModel
-from parler.utils.compat import transaction_atomic
-from parler.utils.i18n import normalize_language_code, get_language_title, is_multilingual_project
+from parler.utils.compat import transaction_atomic, add_preserved_filters
+from parler.utils.i18n import get_language_title, is_multilingual_project
+from parler.utils.views import get_language_parameter, get_language_tabs
 from parler.utils.template import select_template_name
 
 # Code partially taken from django-hvad
@@ -44,7 +46,10 @@ class BaseTranslatableAdmin(BaseModelAdmin):
     """
     The shared code between the regular model admin and inline classes.
     """
+    #: The form to use for the model.
     form = TranslatableModelForm
+
+    #: The URL parameter for the language value.
     query_language_key = 'language'
 
 
@@ -70,26 +75,7 @@ class BaseTranslatableAdmin(BaseModelAdmin):
         """
         Get the language parameter from the current request.
         """
-        if not is_multilingual_project() or not self._has_translatable_model():
-            # By default, the objects are stored in a single static language.
-            # This makes the transition to multilingual easier as well.
-            # The default language can operate as fallback language too.
-            return appsettings.PARLER_DEFAULT_LANGUAGE_CODE
-        else:
-            # In multilingual mode, take the provided language of the request.
-            code = request.GET.get(self.query_language_key)
-
-            if not code:
-                # Show first tab by default
-                try:
-                    lang_choices = appsettings.PARLER_LANGUAGES[settings.SITE_ID]
-                    code = lang_choices[0]['code']
-                except (KeyError, IndexError):
-                    # No configuration, always fallback to default language.
-                    # This is essentially a non-multilingual configuration.
-                    code = appsettings.PARLER_DEFAULT_LANGUAGE_CODE
-
-            return normalize_language_code(code)
+        return get_language_parameter(request, self.query_language_key, object=obj)
 
 
     def get_form_language(self, request, obj=None):
@@ -102,6 +88,19 @@ class BaseTranslatableAdmin(BaseModelAdmin):
             return self._language(request)
 
 
+    def get_queryset_language(self, request):
+        """
+        Return the language to use in the queryset.
+        """
+        if not is_multilingual_project():
+            # Make sure the current translations remain visible, not the dynamically set get_language() value.
+            return appsettings.PARLER_LANGUAGES.get_default_language()
+        else:
+            # Allow to adjust to current language
+            # This is overwritten for the inlines, which follow the primary object.
+            return get_language()
+
+
     def queryset(self, request):
         """
         Make sure the current language is selected.
@@ -112,16 +111,23 @@ class BaseTranslatableAdmin(BaseModelAdmin):
             if not isinstance(qs, TranslatableQuerySet):
                 raise ImproperlyConfigured("{0} class does not inherit from TranslatableQuerySet".format(qs.__class__.__name__))
 
-            if not is_multilingual_project():
-                # Make sure the current translations remain visible, not the dynamically set get_language() value.
-                qs = qs.language(appsettings.PARLER_DEFAULT_LANGUAGE_CODE)
-            else:
-                # Set the initial language for fetched objects.
-                # This is needed for the TranslatableInlineModelAdmin
-                qs = qs.language(self._language(request))
+            # Apply a consistent language to all objects.
+            qs_language = self.get_queryset_language(request)
+            if qs_language:
+                qs = qs.language(qs_language)
+
         return qs
 
 
+    def get_language_tabs(self, request, obj, available_languages, css_class=None):
+        """
+        Determine the language tabs to show.
+        """
+        current_language = self.get_form_language(request, obj)
+        return get_language_tabs(request, current_language, available_languages, css_class=css_class)
+
+
+
 class TranslatableAdmin(BaseTranslatableAdmin, admin.ModelAdmin):
     """
     Base class for translated admins.
@@ -139,7 +145,11 @@ class TranslatableAdmin(BaseTranslatableAdmin, admin.ModelAdmin):
 
     @property
     def change_form_template(self):
-        # Dynamic property to support transition to regular models.
+        """
+        Dynamic property to support transition to regular models.
+
+        This automatically picks ``admin/parler/change_form.html`` when the admin uses a translatable model.
+        """
         if self._has_translatable_model():
             # While this breaks the admin template name detection,
             # the get_change_form_base_template() makes sure it inherits from your template.
@@ -154,7 +164,7 @@ class TranslatableAdmin(BaseTranslatableAdmin, admin.ModelAdmin):
         """
         languages = self.get_available_languages(object)
         languages = [self.get_language_short_title(code) for code in languages]
-        return u'<span class="available-languages">{0}</span>'.format(' '.join(languages))
+        return '<span class="available-languages">{0}</span>'.format(' '.join(languages))
 
     language_column.allow_tags = True
     language_column.short_description = _("Languages")
@@ -174,7 +184,7 @@ class TranslatableAdmin(BaseTranslatableAdmin, admin.ModelAdmin):
         if obj:
             return obj.get_available_languages()
         else:
-            return self.model._translations_model.objects.get_empty_query_set()
+            return self.model._translations_model.objects.none()
 
 
     def get_object(self, request, object_id):
@@ -226,17 +236,19 @@ class TranslatableAdmin(BaseTranslatableAdmin, admin.ModelAdmin):
             lang = get_language_title(lang_code)
 
             available_languages = self.get_available_languages(obj)
-            current_is_translated = lang_code in available_languages
             language_tabs = self.get_language_tabs(request, obj, available_languages)
-
-            context['current_is_translated'] = current_is_translated
-            context['allow_deletion'] = len(available_languages) > 1
             context['language_tabs'] = language_tabs
             if language_tabs:
                 context['title'] = '%s (%s)' % (context['title'], lang)
-            if not current_is_translated:
+            if not language_tabs.current_is_translated:
                 add = True  # lets prepopulated_fields_js work.
 
+            # Patch form_url to contain the "language" GET parameter.
+            # Otherwise AdminModel.render_change_form will clean the URL
+            # and remove the "language" when coming from a filtered object
+            # list causing the wrong translation to be changed.
+            form_url = add_preserved_filters({'preserved_filters': urlencode({'language': lang_code}), 'opts': self.model._meta}, form_url)
+
         # django-fluent-pages uses the same technique
         if 'default_change_form_template' not in context:
             context['default_change_form_template'] = self.get_change_form_base_template()
@@ -278,50 +290,6 @@ class TranslatableAdmin(BaseTranslatableAdmin, admin.ModelAdmin):
         return redirect
 
 
-    def get_language_tabs(self, request, obj, available_languages):
-        """
-        Determine the language tabs to show.
-        """
-        tabs = []
-        get = request.GET.copy()  # QueryDict object
-        language = self.get_form_language(request, obj)
-        tab_languages = []
-
-        base_url = '{0}://{1}{2}'.format(request.is_secure() and 'https' or 'http', request.get_host(), request.path)
-
-        for lang_dict in appsettings.PARLER_LANGUAGES.get(settings.SITE_ID, ()):
-            code = lang_dict['code']
-            title = get_language_title(code)
-            get['language'] = code
-            url = '{0}?{1}'.format(base_url, get.urlencode())
-
-            if code == language:
-                status = 'current'
-            elif code in available_languages:
-                status = 'available'
-            else:
-                status = 'empty'
-
-            tabs.append((url, title, code, status))
-            tab_languages.append(code)
-
-        # Additional stale translations in the database?
-        if appsettings.PARLER_SHOW_EXCLUDED_LANGUAGE_TABS:
-            for code in available_languages:
-                if code not in tab_languages:
-                    get['language'] = code
-                    url = '{0}?{1}'.format(base_url, get.urlencode())
-
-                    if code == language:
-                        status = 'current'
-                    else:
-                        status = 'available'
-
-                    tabs.append((url, get_language_title(code), code, status))
-
-        return tabs
-
-
     @csrf_protect_m
     @transaction_atomic
     def delete_translation(self, request, object_id, language_code):
@@ -362,12 +330,12 @@ class TranslatableAdmin(BaseTranslatableAdmin, admin.ModelAdmin):
         if request.POST: # The user has already confirmed the deletion.
             if perms_needed:
                 raise PermissionDenied
-            obj_display = _('{0} translation of {1}').format(lang, force_unicode(translation))  # in hvad: (translation.master)
+            obj_display = _('{0} translation of {1}').format(lang, force_text(translation))  # in hvad: (translation.master)
 
             self.log_deletion(request, translation, obj_display)
             self.delete_model_translation(request, translation)
             self.message_user(request, _('The %(name)s "%(obj)s" was deleted successfully.') % dict(
-                name=force_unicode(opts.verbose_name), obj=force_unicode(obj_display)
+                name=force_text(opts.verbose_name), obj=force_text(obj_display)
             ))
 
             if self.has_change_permission(request, None):
@@ -375,7 +343,7 @@ class TranslatableAdmin(BaseTranslatableAdmin, admin.ModelAdmin):
             else:
                 return HttpResponseRedirect(reverse('admin:index'))
 
-        object_name = _('{0} Translation').format(force_unicode(opts.verbose_name))
+        object_name = _('{0} Translation').format(force_text(opts.verbose_name))
         if perms_needed or protected:
             title = _("Cannot delete %(name)s") % {"name": object_name}
         else:
@@ -410,7 +378,7 @@ class TranslatableAdmin(BaseTranslatableAdmin, admin.ModelAdmin):
             'opts': opts,
             'app_label': opts.app_label,
             'language_name': get_language_title(language_code),
-            'object_name': force_unicode(opts.verbose_name)
+            'object_name': force_text(opts.verbose_name)
         }
         return render(request, self.deletion_not_allowed_template, context)
 
@@ -432,19 +400,28 @@ class TranslatableAdmin(BaseTranslatableAdmin, admin.ModelAdmin):
         """
         Fetch the inline translations
         """
-        for inline in self.get_inline_instances(request, obj=obj):
+        # django 1.4 do not accept the obj parameter
+        if django.VERSION < (1, 5):
+            inline_instances = self.get_inline_instances(request)
+        else:
+            inline_instances = self.get_inline_instances(request, obj=obj)
+
+        for inline in inline_instances:
             if issubclass(inline.model, TranslatableModel):
                 # leverage inlineformset_factory() to find the ForeignKey.
                 # This also resolves the fk_name if it's set.
                 fk = inline.get_formset(request, obj).fk
 
-                rel_name = 'master__{}'.format(fk.name)
+                rel_name = 'master__{0}'.format(fk.name)
                 filters = {
                     'language_code': language_code,
                     rel_name: obj
                 }
 
                 qs = inline.model._translations_model.objects.filter(**filters)
+                if obj is not None:
+                    qs = qs.using(obj._state.db)
+
                 yield inline, qs
 
 
@@ -461,36 +438,103 @@ class TranslatableAdmin(BaseTranslatableAdmin, admin.ModelAdmin):
         ))
 
 
-_lazy_select_template_name = lazy(select_template_name, unicode)
-
-
-class TranslatableBaseInlineFormSet(BaseInlineFormSet):
-    language_code = None
-
-    def _construct_form(self, i, **kwargs):
-        form = super(TranslatableBaseInlineFormSet, self)._construct_form(i, **kwargs)
-        form.language_code = self.language_code   # Pass the language code for new objects!
-        return form
-
-    def save_new(self, form, commit=True):
-        obj = super(TranslatableBaseInlineFormSet, self).save_new(form, commit)
-        return obj
+_lazy_select_template_name = lazy(select_template_name, six.text_type)
 
 
 class TranslatableInlineModelAdmin(BaseTranslatableAdmin, InlineModelAdmin):
+    """
+    Base class for inline models.
+    """
+    #: The form to use.
     form = TranslatableModelForm
+    #: The formset to use.
     formset = TranslatableBaseInlineFormSet
 
+    @property
+    def inline_tabs(self):
+        """
+        Whether to show inline tabs, can be set as attribute on the inline.
+        """
+        return not self._has_translatable_parent_model()
+
+    def _has_translatable_parent_model(self):
+        # Allow fallback to regular models when needed.
+        return issubclass(self.parent_model, TranslatableModel)
+
+    def get_queryset_language(self, request):
+        if not is_multilingual_project():
+            # Make sure the current translations remain visible, not the dynamically set get_language() value.
+            return appsettings.PARLER_LANGUAGES.get_default_language()
+        else:
+            # Set the initial language for fetched objects.
+            # This is needed for the TranslatableInlineModelAdmin
+            return self._language(request)
+
     def get_formset(self, request, obj=None, **kwargs):
+        """
+        Return the formset, and provide the language information to the formset.
+        """
         FormSet = super(TranslatableInlineModelAdmin, self).get_formset(request, obj, **kwargs)
         # Existing objects already got the language code from the queryset().language() method.
         # For new objects, the language code should be set here.
         FormSet.language_code = self.get_form_language(request, obj)
+
+        if self.inline_tabs:
+            # Need to pass information to the template, this can only happen via the FormSet object.
+            available_languages = self.get_available_languages(obj, FormSet)
+            FormSet.language_tabs = self.get_language_tabs(request, obj, available_languages, css_class='parler-inline-language-tabs')
+            FormSet.language_tabs.allow_deletion = self._has_translatable_parent_model()   # Views not available otherwise.
+
         return FormSet
 
+    def get_form_language(self, request, obj=None):
+        """
+        Return the current language for the currently displayed object fields.
+        """
+        if self._has_translatable_parent_model():
+            return super(TranslatableInlineModelAdmin, self).get_form_language(request, obj=obj)
+        else:
+            # Follow the ?language parameter
+            return self._language(request)
+
+    def get_available_languages(self, obj, formset):
+        """
+        Fetching the available inline languages as queryset.
+        """
+        if obj:
+            # Inlines dictate language code, not the parent model.
+            # Hence, not looking at obj.get_available_languages(), but see what languages
+            # are used by the inline objects that point to it.
+            filter = {
+                'master__{0}'.format(formset.fk.name): obj
+            }
+            return self.model._translations_model.objects.using(obj._state.db).filter(**filter) \
+                   .values_list('language_code', flat=True).distinct().order_by('language_code')
+        else:
+            return self.model._translations_model.objects.none()
+
 
 class TranslatableStackedInline(TranslatableInlineModelAdmin):
-    template = 'admin/edit_inline/stacked.html'
+    """
+    The inline class for stacked layout.
+    """
+    @property
+    def template(self):
+        if self.inline_tabs:
+            return 'admin/parler/edit_inline/stacked_tabs.html'
+        else:
+            # Admin default
+            return 'admin/edit_inline/stacked.html'
+
 
 class TranslatableTabularInline(TranslatableInlineModelAdmin):
-    template = 'admin/edit_inline/tabular.html'
+    """
+    The inline class for tabular layout.
+    """
+    @property
+    def template(self):
+        if self.inline_tabs:
+            return 'admin/parler/edit_inline/tabular_tabs.html'
+        else:
+            # Admin default
+            return 'admin/edit_inline/tabular.html'
