@@ -1,27 +1,15 @@
 from django import forms
-from django.forms.models import ModelFormMetaclass
+from django.forms.models import ModelFormMetaclass, BaseInlineFormSet
 from django.utils.translation import get_language
+from django.utils import six
 from parler.models import TranslatableModel, TranslationDoesNotExist
 
-
-def get_model_form_field(model, name, formfield_callback=None, **kwargs):
-    """
-    Utility to create the formfield from a model field.
-    When a field is not editable, a ``None`` will be returned.
-    """
-    field = model._meta.get_field_by_name(name)[0]
-    if not field.editable:  # see fields_for_model() logic in Django.
-        return None
-
-    # Apply admin formfield_overrides
-    if formfield_callback is None:
-        formfield = field.formfield(**kwargs)
-    elif not callable(formfield_callback):
-        raise TypeError('formfield_callback must be a function or callable')
-    else:
-        formfield = formfield_callback(field, **kwargs)
-
-    return formfield
+__all__ = (
+    'TranslatableModelForm',
+    'TranslatedField',
+    'TranslatableModelFormMixin',
+    #'TranslatableModelFormMetaclass',
+)
 
 
 class TranslatedField(object):
@@ -46,12 +34,13 @@ class TranslatedField(object):
 
 class TranslatableModelFormMixin(object):
     """
-    Form mixin, to fetch+store translated fields.
+    The base methods added to :class:`TranslatableModelForm` to fetch and store translated fields.
     """
-    language_code = None    # Set by TranslatableAdmin.get_form()
+    language_code = None    # Set by TranslatableAdmin.get_form() on the constructed subclass.
 
 
     def __init__(self, *args, **kwargs):
+        current_language = kwargs.pop('_current_language', None)   # Used for TranslatableViewMixin
         super(TranslatableModelFormMixin, self).__init__(*args, **kwargs)
 
         # Load the initial values for the translated fields
@@ -70,18 +59,29 @@ class TranslatableModelFormMixin(object):
 
         # Typically already set by admin
         if self.language_code is None:
-            self.language_code = instance.get_current_language() if instance is not None else get_language()
+            if instance:
+                self.language_code = instance.get_current_language()
+                return
+            else:
+                self.language_code = current_language or get_language()
 
+        # Make sure the language code is set as early as possible (so it's active during clean() methods)
+        #self.instance.set_current_language(self.language_code)
 
-    def save(self, commit=True):
-        # Assign translated fields to the model (using the TranslatedAttribute descriptor)
+    def save(self, *args, **kwargs):
+        # Using args, kwargs to support custom parent arguments too.
         self.instance.set_current_language(self.language_code)
+        self.save_translated_fields(*args, **kwargs)
+        return super(TranslatableModelFormMixin, self).save(*args, **kwargs)
+
+    def save_translated_fields(self, *args, **kwargs):
+        """
+        Save all translated fields.
+        """
+        # Assign translated fields to the model (using the TranslatedAttribute descriptor)
         for field in self._get_translated_fields():
             setattr(self.instance, field, self.cleaned_data[field])
 
-        return super(TranslatableModelFormMixin, self).save(commit)
-
-
     def _get_translated_fields(self):
         return [f_name for f_name in self._meta.model._translations_model.get_translated_fields() if f_name in self.fields]
 
@@ -107,7 +107,7 @@ class TranslatableModelFormMetaclass(ModelFormMetaclass):
 
             # Detect all placeholders at this class level.
             translated_fields = [
-                f_name for f_name, attr_value in attrs.iteritems() if isinstance(attr_value, TranslatedField)
+                f_name for f_name, attr_value in six.iteritems(attrs) if isinstance(attr_value, TranslatedField)
             ]
 
             # Include the translated fields as attributes, pretend that these exist on the form.
@@ -115,28 +115,39 @@ class TranslatableModelFormMetaclass(ModelFormMetaclass):
             # since the admin always uses modelform_factory() on the form class, and therefore triggering this metaclass.
             if form_model:
                 translations_model = form_model._translations_model
+                fields = getattr(form_new_meta, 'fields', form_meta.fields)
                 exclude = getattr(form_new_meta, 'exclude', form_meta.exclude) or ()
                 widgets = getattr(form_new_meta, 'widgets', form_meta.widgets) or ()
                 formfield_callback = attrs.get('formfield_callback', None)
 
+                if fields == '__all__':
+                    fields = None
+
                 for f_name in translations_model.get_translated_fields():
                     # Add translated field if not already added, and respect exclude options.
                     if f_name in translated_fields:
                         # The TranslatedField placeholder can be replaced directly with actual field, so do that.
-                        attrs[f_name] = get_model_form_field(translations_model, f_name, formfield_callback=formfield_callback, **translated_fields[f_name].kwargs)
-                    elif f_name not in form_base_fields and not f_name in attrs and f_name not in exclude:
-                        # See if this formfield was previously defined using a TranslatedField placeholder.
-                        placeholder = _get_mro_attribute(bases, f_name)
-                        if placeholder and isinstance(placeholder, TranslatedField):
-                            kwargs = placeholder.kwargs
-                        elif f_name in widgets:
+                        attrs[f_name] = _get_model_form_field(translations_model, f_name, formfield_callback=formfield_callback, **translated_fields[f_name].kwargs)
+                    # The next code holds the same logic as fields_for_model()
+                    # The f.editable check happens in _get_model_form_field()
+                    elif f_name not in form_base_fields \
+                     and (fields is None or f_name in fields) \
+                     and f_name not in exclude \
+                     and not f_name in attrs:
+                        # Get declared widget kwargs
+                        if f_name in widgets:
                             # Not combined with declared fields (e.g. the TranslatedField placeholder)
                             kwargs = {'widget': widgets[f_name]}
                         else:
                             kwargs = {}
 
+                        # See if this formfield was previously defined using a TranslatedField placeholder.
+                        placeholder = _get_mro_attribute(bases, f_name)
+                        if placeholder and isinstance(placeholder, TranslatedField):
+                            kwargs.update(placeholder.kwargs)
+
                         # Add the form field as attribute to the class.
-                        formfield = get_model_form_field(translations_model, f_name, formfield_callback=formfield_callback, **kwargs)
+                        formfield = _get_model_form_field(translations_model, f_name, formfield_callback=formfield_callback, **kwargs)
                         if formfield is not None:
                             attrs[f_name] = formfield
 
@@ -154,8 +165,46 @@ def _get_mro_attribute(bases, name, default=None):
     return default
 
 
-class TranslatableModelForm(TranslatableModelFormMixin, forms.ModelForm):
+
+def _get_model_form_field(model, name, formfield_callback=None, **kwargs):
     """
-    A model form for translated models.
+    Utility to create the formfield from a model field.
+    When a field is not editable, a ``None`` will be returned.
     """
-    __metaclass__ = TranslatableModelFormMetaclass
+    field = model._meta.get_field_by_name(name)[0]
+    if not field.editable:  # see fields_for_model() logic in Django.
+        return None
+
+    # Apply admin formfield_overrides
+    if formfield_callback is None:
+        formfield = field.formfield(**kwargs)
+    elif not callable(formfield_callback):
+        raise TypeError('formfield_callback must be a function or callable')
+    else:
+        formfield = formfield_callback(field, **kwargs)
+
+    return formfield
+
+
+class TranslatableModelForm(six.with_metaclass(TranslatableModelFormMetaclass, TranslatableModelFormMixin), forms.ModelForm):
+    """
+    The model form to use for translated models.
+    """
+    # six.with_metaclass does not handle more than 2 parent classes for django < 1.6
+    # so only one is wrapped within with_metaclass
+
+
+class TranslatableBaseInlineFormSet(BaseInlineFormSet):
+    """
+    The formset base for creating inlines with translatable models.
+    """
+    language_code = None
+
+    def _construct_form(self, i, **kwargs):
+        form = super(TranslatableBaseInlineFormSet, self)._construct_form(i, **kwargs)
+        form.language_code = self.language_code   # Pass the language code for new objects!
+        return form
+
+    def save_new(self, form, commit=True):
+        obj = super(TranslatableBaseInlineFormSet, self).save_new(form, commit)
+        return obj
