@@ -1,22 +1,70 @@
 """
-Simple but effective translation support.
+The models and fields for translation support.
 
-Integrating *django-hvad* (v0.3) in advanced projects turned out to be really hard,
-as it changes the behavior of the QuerySet iterator, manager methods
-and model metaclass which *django-polymorphic* and friends also rely on.
-The following is a "crude, but effective" way to introduce multilingual support.
+The default is to use the :class:`TranslatedFields` class in the model, like:
 
-Added on top of that, the API-suger is provided, similar to what django-hvad has.
-It's possible to create the translations model manually,
-or let it be created dynamically when using the :class:`TranslatedFields` field.
+.. code-block:: python
+
+    from django.db import models
+    from parler.models import TranslatableModel, TranslatedFields
+
+
+    class MyModel(TranslatableModel):
+        translations = TranslatedFields(
+            title = models.CharField(_("Title"), max_length=200)
+        )
+
+        class Meta:
+            verbose_name = _("MyModel")
+
+        def __unicode__(self):
+            return self.title
+
+
+It's also possible to create the translated fields model manually:
+
+.. code-block:: python
+
+    from django.db import models
+    from parler.models import TranslatableModel, TranslatedFieldsModel
+    from parler.fields import TranslatedField
+
+
+    class MyModel(TranslatableModel):
+        title = TranslatedField()  # Optional, explicitly mention the field
+
+        class Meta:
+            verbose_name = _("MyModel")
+
+        def __unicode__(self):
+            return self.title
+
+
+    class MyModelTranslation(TranslatedFieldsModel):
+        master = models.ForeignKey(MyModel, related_name='translations', null=True)
+        title = models.CharField(_("Title"), max_length=200)
+
+        class Meta:
+            verbose_name = _("MyModel translation")
+
+This has the same effect, but also allows to to override
+the :func:`~django.db.models.Model.save` method, or add new methods yourself.
+
+The translated model is compatible with django-hvad, making the transition between both projects relatively easy.
+The manager and queryset objects of django-parler can work together with django-mptt and django-polymorphic.
 """
+from __future__ import unicode_literals
+import django
+from django.conf import settings
 from django.core.exceptions import ImproperlyConfigured
 from django.db import models, router
 from django.db.models.base import ModelBase
 from django.db.models.fields.related import ReverseSingleRelatedObjectDescriptor
 from django.utils.functional import lazy
-from django.utils.translation import get_language, ugettext
+from django.utils.translation import get_language, ugettext, ugettext_lazy as _
+from django.utils import six
 from parler import signals
+from parler.cache import _cache_translation, _cache_translation_needs_fallback, _delete_cached_translation, get_cached_translation, _delete_cached_translations, get_cached_translated_field
 from parler.fields import TranslatedField, LanguageCodeDescriptor, TranslatedFieldDescriptor
 from parler.managers import TranslatableManager
 from parler.utils.i18n import normalize_language_code, get_language_settings, get_language_title
@@ -26,17 +74,34 @@ import logging
 logger = logging.getLogger(__name__)
 
 
+__all__ = (
+    'TranslatableModel',
+    'TranslatedFields',
+    'TranslatedFieldsModel',
+    'TranslatedFieldsModelBase',
+    'TranslationDoesNotExist',
+    #'create_translations_model',
+)
+
 
 class TranslationDoesNotExist(AttributeError):
     """
     A tagging interface to detect missing translations.
-    The exception inherits from :class:`AttributeError` to reflect what is actually happening.
-    It also causes the templates to handle the missing attributes silently, which is very useful in the admin for example.
+    The exception inherits from :class:`~exceptions.AttributeError` to reflect what is actually happening.
+    Therefore it also causes the templates to handle the missing attributes silently, which is very useful in the admin for example.
+
+    This class is also used in the ``DoesNotExist`` object on the translated model, which inherits from:
+
+    * this class
+    * the ``sharedmodel.DoesNotExist`` class
+    * the original ``translatedmodel.DoesNotExist`` class.
+
+    This makes sure that the regular code flow is decently handled by existing exception handlers.
     """
     pass
 
 
-_lazy_verbose_name = lazy(lambda x: ugettext("{0} Translation").format(x._meta.verbose_name), unicode)
+_lazy_verbose_name = lazy(lambda x: ugettext("{0} Translation").format(x._meta.verbose_name), six.text_type)
 
 
 def create_translations_model(shared_model, related_name, meta, **fields):
@@ -65,11 +130,11 @@ def create_translations_model(shared_model, related_name, meta, **fields):
     meta.setdefault('verbose_name', _lazy_verbose_name(shared_model))
 
     # Define attributes for translation table
-    name = '{0}_Translation'.format(shared_model.__name__)
+    name = str('{0}Translation'.format(shared_model.__name__))  # makes it bytes, for type()
 
     attrs = {}
     attrs.update(fields)
-    attrs['Meta'] = type('Meta', (object,), meta)
+    attrs['Meta'] = type(str('Meta'), (object,), meta)
     attrs['__module__'] = shared_model.__module__
     attrs['objects'] = models.Manager()
     attrs['master'] = models.ForeignKey(shared_model, related_name=related_name, editable=False, null=True)
@@ -92,7 +157,10 @@ class TranslatedFields(object):
 
     The field name becomes the related name of the :class:`TranslatedFieldsModel` subclass.
 
-    Example::
+    Example:
+
+    .. code-block:: python
+
         from django.db import models
         from parler.models import TranslatableModel, TranslatedFields
 
@@ -121,6 +189,8 @@ class TranslatedFields(object):
 class TranslatableModel(models.Model):
     """
     Base model class to handle translations.
+
+    All translatable fields will appear on this model, proxying the calls to the :class:`TranslatedFieldsModel`.
     """
 
     # Consider these fields "protected" or "internal" attributes.
@@ -158,7 +228,7 @@ class TranslatableModel(models.Model):
         # Assign translated args manually.
         if translated_kwargs:
             translation = self._get_translated_model(auto_create=True)
-            for field, value in translated_kwargs.iteritems():
+            for field, value in six.iteritems(translated_kwargs):
                 setattr(translation, field, value)
 
 
@@ -199,7 +269,7 @@ class TranslatableModel(models.Model):
             language_code = self._current_language
 
         try:
-            # Check the cache directly, and the answer is known.
+            # Check the local cache directly, and the answer is known.
             # NOTE this may also return newly auto created translations which are not saved yet.
             return self._translations_cache[language_code] is not None
         except KeyError:
@@ -216,7 +286,11 @@ class TranslatableModel(models.Model):
         """
         Return the language codes of all translated variations.
         """
-        return self._translations_model.objects.filter(master=self).values_list('language_code', flat=True).order_by('language_code')
+        qs = self._get_translated_queryset()
+        if qs._prefetch_done:
+            return sorted(obj.language_code for obj in qs)
+        else:
+            return qs.values_list('language_code', flat=True).order_by('language_code')
 
 
     def _get_translated_model(self, language_code=None, use_fallback=False, auto_create=False):
@@ -229,7 +303,7 @@ class TranslatableModel(models.Model):
         if not language_code:
             language_code = self._current_language
 
-        # 1. fetch the object from the cache
+        # 1. fetch the object from the local cache
         try:
             object = self._translations_cache[language_code]
 
@@ -238,15 +312,37 @@ class TranslatableModel(models.Model):
                 return object
         except KeyError:
             # 2. No cache, need to query
-            # Get via self.TRANSLATIONS_FIELD.get(..) so it also uses the prefetch/select_related cache.
-            accessor = getattr(self, self._translations_field)
-            try:
-                object = accessor.get(language_code=language_code)
-            except self._translations_model.DoesNotExist:
-                pass
-            else:
-                self._translations_cache[language_code] = object
-                return object
+            # Check that this object already exists, would be pointless otherwise to check for a translation.
+            if not self._state.adding and self.pk:
+                qs = self._get_translated_queryset()
+                if qs._prefetch_done:
+                    # 2.1, use prefetched data
+                    # If the object is not found in the prefetched data (which contains all translations),
+                    # it's pointless to check for memcached (2.2) or perform a single query (2.3)
+                    for object in qs:
+                        if object.language_code == language_code:
+                            self._translations_cache[language_code] = object
+                            _cache_translation(object)  # Store in memcached
+                            return object
+                else:
+                    # 2.2, fetch from memcached
+                    object = get_cached_translation(self, language_code, use_fallback=use_fallback)
+                    if object is not None:
+                        # Track in local cache
+                        if object.language_code != language_code:
+                            self._translations_cache[language_code] = None  # Set fallback marker
+                        self._translations_cache[object.language_code] = object
+                        return object
+                    else:
+                        # 2.3, fetch from database
+                        try:
+                            object = qs.get(language_code=language_code)
+                        except self._translations_model.DoesNotExist:
+                            pass
+                        else:
+                            self._translations_cache[language_code] = object
+                            _cache_translation(object)  # Store in memcached
+                            return object
 
         # Not in cache, or default.
         # Not fetched from DB
@@ -259,6 +355,7 @@ class TranslatableModel(models.Model):
                 master=self  # ID might be None at this point
             )
             self._translations_cache[language_code] = object
+            # Not stored in memcached here yet, first fill + save it.
             return object
 
         # 4. Fallback?
@@ -266,18 +363,23 @@ class TranslatableModel(models.Model):
         lang_dict = get_language_settings(language_code)
 
         if use_fallback and (lang_dict['fallback'] != language_code):
+            # Explicitly set a marker for the fact that this translation uses the fallback instead.
+            # Avoid making that query again.
+            self._translations_cache[language_code] = None  # None value is the marker.
+            if not self._state.adding or self.pk:
+                _cache_translation_needs_fallback(self, language_code)
+
             # Jump to fallback language, return directly.
             # Don't cache under this language_code
-            self._translations_cache[language_code] = None   # explicit marker that language query was tried before.
             try:
                 return self._get_translated_model(lang_dict['fallback'], use_fallback=False, auto_create=auto_create)
             except self._translations_model.DoesNotExist:
-                fallback_msg = u" (tried fallback {0})".format(lang_dict['fallback'])
+                fallback_msg = " (tried fallback {0})".format(lang_dict['fallback'])
 
         # None of the above, bail out!
         raise self._translations_model.DoesNotExist(
-            u"{0} does not have a translation for the current language!\n"
-            u"{0} ID #{1}, language={2}{3}".format(self._meta.verbose_name, self.pk, language_code, fallback_msg or ''
+            "{0} does not have a translation for the current language!\n"
+            "{0} ID #{1}, language={2}{3}".format(self._meta.verbose_name, self.pk, language_code, fallback_msg or ''
         ))
 
 
@@ -292,45 +394,90 @@ class TranslatableModel(models.Model):
             try:
                 return self._translations_cache.get(self._current_language, None) \
                     or self._translations_cache.get(self.get_fallback_language(), None) \
-                    or next(t for t in self._translations_cache.itervalues() if t if not None)  # Skip fallback markers.
+                    or next(t for t in six.itervalues(self._translations_cache) if t if not None)  # Skip fallback markers.
             except StopIteration:
                 pass
 
         try:
-            translation = self._translations_model.objects.filter(master=self)[0]
+            # Use prefetch if available, otherwise perform separate query.
+            qs = self._get_translated_queryset()
+            if qs._prefetch_done:
+                translation = list(qs)[0]
+            else:
+                translation = qs[0]
         except IndexError:
             return None
         else:
             self._translations_cache[translation.language_code] = translation
+            _cache_translation(translation)
             return translation
 
 
+    def _get_translated_queryset(self):
+        """
+        Return the queryset that points to the translated model.
+        If there is a prefetch, it can be read from this queryset.
+        """
+        # Get via self.TRANSLATIONS_FIELD.get(..) so it also uses the prefetch/select_related cache.
+        accessor = getattr(self, self._translations_field)
+        try:
+            return accessor.get_queryset()
+        except AttributeError:
+            # Fallback for Django 1.4 and Django 1.5
+            return accessor.get_query_set()
+
+
     def save(self, *args, **kwargs):
         super(TranslatableModel, self).save(*args, **kwargs)
-        self.save_translations()
+        self.save_translations(*args, **kwargs)
 
 
-    def save_translations(self):
+    def delete(self, using=None):
+        _delete_cached_translations(self)
+        super(TranslatableModel, self).delete(using)
+
+
+    def save_translations(self, *args, **kwargs):
+        """
+        The method to save all translations.
+        This can be overwritten to implement any custom additions.
+        This method calls :func:`save_translation` for every fetched language.
+
+        :param args: Any custom arguments to pass to :func:`save`.
+        :param kwargs: Any custom arguments to pass to :func:`save`.
+        """
+        # Copy cache, new objects (e.g. fallbacks) might be fetched if users override save_translation()
+        translations = self._translations_cache.values()
+
         # Save all translated objects which were fetched.
         # This also supports switching languages several times, and save everything in the end.
-        for translation in self._translations_cache.itervalues():
+        for translation in translations:
             if translation is None:  # Skip fallback markers
                 continue
 
-            self.save_translation(translation)
+            self.save_translation(translation, *args, **kwargs)
 
 
-    def save_translation(self, translation):
+    def save_translation(self, translation, *args, **kwargs):
+        """
+        Save the translation when it's modified, or unsaved.
+
+        :param translation: The translation
+        :type translation: TranslatedFieldsModel
+        :param args: Any custom arguments to pass to :func:`save`.
+        :param kwargs: Any custom arguments to pass to :func:`save`.
+        """
         # Translation models without any fields are also supported.
         # This is useful for parent objects that have inlines;
         # the parent object defines how many translations there are.
         if translation.is_modified or (translation.is_empty and not translation.pk):
             if not translation.master_id:  # Might not exist during first construction
+                translation._state.db = self._state.db
                 translation.master = self
-            translation.save()
+            translation.save(*args, **kwargs)
 
 
-    def safe_translation_getter(self, field, default=None, any_language=False):
+    def safe_translation_getter(self, field, default=None, language_code=None, any_language=False):
         """
         Fetch a translated property, and return a default value
         when both the translation and fallback language are missing.
@@ -341,8 +488,24 @@ class TranslatableModel(models.Model):
         Also consider using ``field = TranslatedField(any_language=True)`` in the model itself,
         to make this behavior the default for the given field.
         """
+        # By default, query via descriptor (TranslatedFieldDescriptor)
+        # which also attempts the fallback language if configured to do so.
+        tr_model = self
+
+        # Extra feature: query a single field from a other translation.
+        if language_code and language_code != self._current_language:
+            # Try to fetch a cached value first.
+            value = get_cached_translated_field(self, language_code, field)
+            if value is not None:
+                return value
+
+            try:
+                tr_model = self._get_translated_model(language_code)
+            except TranslationDoesNotExist:
+                pass  # Use 'self'
+
         try:
-            return getattr(self, field)
+            return getattr(tr_model, field)
         except TranslationDoesNotExist:
             pass
 
@@ -359,11 +522,22 @@ class TranslatedFieldsModelBase(ModelBase):
     Meta-class for the translated fields model.
 
     It performs the following steps:
-    - It validates the 'master' field, in case it's added manually.
-    - It tells the original model to use this model for translations.
-    - It adds the proxy attributes to the shared model.
+
+    * It validates the 'master' field, in case it's added manually.
+    * It tells the original model to use this model for translations.
+    * It adds the proxy attributes to the shared model.
     """
     def __new__(mcs, name, bases, attrs):
+
+        # Workaround compatibility issue with six.with_metaclass() and custom Django model metaclasses:
+        if not attrs and name == 'NewBase':
+            if django.VERSION < (1,5):
+                # Let Django fully ignore the class which is inserted in between.
+                # Django 1.5 fixed this, see https://code.djangoproject.com/ticket/19688
+                attrs['__module__'] = 'django.utils.six'
+                attrs['Meta'] = type(str('Meta'), (), {'abstract': True})
+            return super(TranslatedFieldsModelBase, mcs).__new__(mcs, name, bases, attrs)
+
         new_class = super(TranslatedFieldsModelBase, mcs).__new__(mcs, name, bases, attrs)
         if bases[0] == models.Model:
             return new_class
@@ -405,16 +579,16 @@ def _validate_master(new_class):
     return shared_model
 
 
-
-class TranslatedFieldsModel(models.Model):
+class TranslatedFieldsModel(six.with_metaclass(TranslatedFieldsModelBase, models.Model)):
     """
     Base class for the model that holds the translated fields.
     """
-    __metaclass__ = TranslatedFieldsModelBase
+    language_code = models.CharField(_("Language"), choices=settings.LANGUAGES, max_length=15, db_index=True)
 
-    language_code = models.CharField(max_length=15, db_index=True)
+    #: The mandatory Foreign key field to the shared model.
     master = None   # FK to shared model.
 
+
     class Meta:
         abstract = True
 
@@ -427,14 +601,23 @@ class TranslatedFieldsModel(models.Model):
 
     @property
     def is_modified(self):
+        """
+        Tell whether the object content is modified since fetching it.
+        """
         return self._original_values != self._get_field_values()
 
     @property
     def is_empty(self):
+        """
+        True when there are no translated fields.
+        """
         return len(self.get_translated_fields()) == 0
 
     @property
     def shared_model(self):
+        """
+        Returns the shared model this model is linked to.
+        """
         return self.__class__.master.field.rel.to
 
     def save_base(self, raw=False, using=None, **kwargs):
@@ -450,6 +633,7 @@ class TranslatedFieldsModel(models.Model):
         # Perform save
         super(TranslatedFieldsModel, self).save_base(raw=raw, using=using, **kwargs)
         self._original_values = self._get_field_values()
+        _cache_translation(self)
 
         # Send the post_save signal
         if not self._meta.auto_created:
@@ -465,6 +649,7 @@ class TranslatedFieldsModel(models.Model):
             signals.pre_translation_delete.send(sender=self.shared_model, instance=self, using=using)
 
         super(TranslatedFieldsModel, self).delete(using=using)
+        _delete_cached_translation(self)
 
         # Send post-delete signal
         if not self._meta.auto_created:
@@ -472,7 +657,7 @@ class TranslatedFieldsModel(models.Model):
 
     def _get_field_values(self):
         # Return all field values in a consistent (sorted) manner.
-        return [getattr(self, field) for field in self._meta.get_all_field_names()]
+        return [getattr(self, field.get_attname()) for field, _ in self._meta.get_fields_with_model()]
 
     @classmethod
     def get_translated_fields(cls):
@@ -493,21 +678,31 @@ class TranslatedFieldsModel(models.Model):
             try:
                 # Check if the field already exists.
                 # Note that the descriptor even proxies this request, so it should return our field.
-                field = getattr(shared_model, name)
+                shared_field = getattr(shared_model, name)
             except AttributeError:
                 # Add the proxy field for the shared field.
                 TranslatedField().contribute_to_class(shared_model, name)
             else:
-                if not isinstance(field, (models.Field, TranslatedFieldDescriptor)):
+                # Currently not allowing to replace existing model fields with translatable fields.
+                # That would be a nice feature addition however.
+                if not isinstance(shared_field, (models.Field, TranslatedFieldDescriptor)):
                     raise TypeError("The model '{0}' already has a field named '{1}'".format(shared_model.__name__, name))
 
+                # When the descriptor was placed on an abstract model,
+                # it doesn't point to the real model that holds the _translations_model
+                # "Upgrade" the descriptor on the class
+                if shared_field.field.model is not shared_model:
+                    TranslatedField(any_language=shared_field.field.any_language).contribute_to_class(shared_model, name)
+
+
         # Make sure the DoesNotExist error can be detected als shared_model.DoesNotExist too,
         # and by inheriting from AttributeError it makes sure (admin) templates can handle the missing attribute.
-        cls.DoesNotExist = type('DoesNotExist', (TranslationDoesNotExist, shared_model.DoesNotExist, cls.DoesNotExist,), {})
-
+        cls.DoesNotExist = type(str('DoesNotExist'), (TranslationDoesNotExist, shared_model.DoesNotExist, cls.DoesNotExist,), {})
 
     def __unicode__(self):
-        return get_language_title(self.language_code)
+        # use format to avoid weird error in django 1.4
+        # TypeError: coercing to Unicode: need string or buffer, __proxy__ found
+        return "{0}".format(get_language_title(self.language_code))
 
     def __repr__(self):
         return "<{0}: #{1}, {2}, master: #{3}>".format(
